; -------------------------------------------------------------------
; This is the MASM 6+ source for a simple 16-bit DOS program that
; installs a temporary handler for interrupt 21h and before loading
; and executing QBasic. The interrupt handler counts the total
; number of function calls and sets a flag for each called function.
; After QBasic terminates the total number of function calls and the
; function number of the called functions are displayed.
;
; This program passes its command line and environment to QBasic.
;
; Note that QBasic must be in the current directory or somewhere
; on the path.
; -------------------------------------------------------------------

displaychar MACRO vararg:VARARG
    FOR arg, <vararg>
      mov dl, arg
      mov ah, 2
      int 21h
    ENDM
ENDM

PROGRAM_NAME  EQU "qbasic.exe"

; Declare the LoadExec structure used by the DOS
; Load and Execute Program function.

LoadExec STRUCT
  leEnvironment dw 0          ; environment-block segment
  leCommandTail dd 0          ; address of command tail
  leFCB_1       dd 0          ; address of default FCB #1
  leFCB_2       dd 0          ; address of default FCB #1
LoadExec ENDS

.model tiny                   ; generate a COM file
.386                          ; enable 32-bit instructions
.stack                        ; default size (will be changed)
.code                         ; start code segmemt
.startup                      ; for tiny, expands to ORG 100h

    ; Place the data in the code segment for easy access from the
    ; interrupt handler (not actually necessary for tiny model).

    jmp @F                    ; jump around data

    LoadArgs   LoadExec <>    ; use default initializers

    prevIsr21h dd 0           ; previous Interrupt 21h handler

    count      dw 0           ; total number of interrupt 21h calls

    flags     db 256 dup(0)   ; one flag for each function

    progName   db PROGRAM_NAME,0

    msgCalls   db "Total number of Interrupt 21h calls: $"
    msgCalled  db "Interrupt 21h functions called:",13,10,"$"
    msgExec    db "Error executing QBasic, $"
    msgExit    db "Press any key to exit...$"

  @@:

    ; Move stack and release unneeded memory.

    mov bx, OFFSET theEnd
    mov sp, bx                ; move stack
    shr bx, 4                 ; convert to paragraphs
    inc bx                    ; correct for remainder
    mov ah, 4Ah               ; set memory block size
    int 21h                   ; (ES specifies memory block)

    ; Hook Interrupt 21h.

    mov ax, 3521h             ; get interrupt vector
    int 21h                   ; returns handler address in ES:BX
    mov WORD PTR prevIsr21h,bx
    mov WORD PTR prevIsr21h+2,es
    mov ax, 2521h             ; set interrupt vector
    mov dx, OFFSET Isr21h     ; DS:DX -> new interrupt handler
    int 21h

    ; Initialize LoadArgs such that child program will receive
    ; exact copy of our environment, commandTail, and FCB's.

    mov LoadArgs.leEnvironment,0
    mov WORD PTR LoadArgs.leCommandTail,80h
    mov WORD PTR LoadArgs.leCommandTail+2, ds
    mov WORD PTR LoadArgs.leFCB_1, 5Ch
    mov WORD PTR LoadArgs.leFCB_1+2, ds
    mov WORD PTR LoadArgs.leFCB_2, 6Ch
    mov WORD PTR LoadArgs.leFCB_2+2, ds

    ; Execute QBasic directly as a child. The EXEC function will
    ; not return until QBasic terminates.

    mov dx, OFFSET progName   ; DS:DX -> path
    push ds
    pop es
    mov bx, OFFSET LoadArgs   ; ES:BX -> LoadExec structure
    mov ax, 4B00h             ; load and execute
    int 21h
    jnc @F
    mov dx, OFFSET msgExec
    mov ah, 9
    int 21h
  @@:

    ; Unhook the interrupt vector.

    push ds
    mov ax, 2521h             ; set interrupt vector
    lds dx, prevIsr21h        ; DS:DX -> interrupt handler
    int 21h                   ; restore original handler
    pop ds

    mov dx, OFFSET msgCalls
    mov ah, 9
    int 21h

    mov bx, count
    mov al, bh
    call displayhex
    mov al, bl
    call displayhex
    displaychar "h",13,10

    mov dx, OFFSET msgCalled
    mov ah, 9
    int 21h

    xor bx, bx
    .WHILE bx < 256
      .IF flags[bx]
        mov al, bl
        call displayhex
        displaychar "h",9
      .ENDIF
      inc bx
    .ENDW
    displaychar 13,10

    mov dx, OFFSET msgExit
    mov ah, 9
    int 21h

    mov ah, 10h               ; get keystroke (will wait)
    int 16h

    .exit

; -------------------------------------------------------------------
; Interrupt 21h handler.
; -------------------------------------------------------------------

Isr21h proc

    push bx                   ; must preserve any register we change
    push cx
    inc cs:count
    mov bx, cs
    mov cx, ds
    .IF bx != cx              ; ignore calls from this program
      movzx bx, ah
      mov cs:flags[bx], 1     ; set flag
    .ENDIF
    pop cx
    pop bx
    jmp cs:prevIsr21h         ; chain to previous handler

Isr21h endp

; -------------------------------------------------------------------

hexChars db "0123456789ABCDEF"
displayhex proc uses ax bx
    ; Display byte in AL as hex.
    push ax
    movzx bx, al
    shr bx, 4
    displaychar hexChars[bx]
    pop ax
    movzx bx, al
    and bx, 0fh
    displaychar hexChars[bx]
    ret
displayhex endp

; -------------------------------------------------------------------
; Allocate a new stack.
; -------------------------------------------------------------------

    even                      ; specify word alignment
    dw 100h dup(0)            ; allocate 256-words

theEnd:                       ; label end of program

; -------------------------------------------------------------------
end
